import re

DECORATOR_GROUP_PATTERN = r"([^\w\d\s\']+)"
DECORATOR_GROUP_REGEX = re.compile(DECORATOR_GROUP_PATTERN)

DOUBLE_QUOTE_PATTERN = r"\""
DOUBLE_QUOTE_REGEX = re.compile(DOUBLE_QUOTE_PATTERN)

QUOTED_TEXT_PATTERN = r"\'([^\"\'\s]+)\'"
QUOTED_TEXT_REGEX = re.compile(QUOTED_TEXT_PATTERN)

MEANINGLESS_PATTERN = r"^[^\w\d\!\$\%\&\?\.]+$"
MEANINGLESS_REGEX = re.compile(MEANINGLESS_PATTERN)